{% load i18n %}

<p class="text-muted">{% translate 'Color code:' %}
    <span class="text-success">{% translate 'fully compliant' %}</span> |
    <span class="text-warning">{% translate 'partly compliant' %}</span> |
    <span class="text-danger">{% translate 'not compliant' %}</span>
</p>
