{% load i18n %}
{% load static %}
{% load humanize %}
{% load evelinks %}

<!-- Modal for Mails -->
<div class="modal" id="modalCharacterMail" tabindex="-1" role="dialog" aria-labelledby="myModalLabel">
    <div class="modal-dialog modal-lg" role="document">
        <div class="modal-content">

            <div class="modal-header">
                <button type="button" class="close" data-dismiss="modal" aria-label="Close"><span aria-hidden="true">
                    &times;
                </span></button>
                <h4 class="modal-title" id="myModalLabel">{% translate 'Mail' %}</h4>
            </div>

            <div class="modal-body">
                <div id="div_mail_body_loader">
                    {% if NIGHT_MODE %}
                        <img src="{% static 'memberaudit/images/Spinner-1s-64px-dark.gif' %}"/>
                    {% else %}
                        <img src="{% static 'memberaudit/images/Spinner-1s-64px-light.gif' %}"/>
                    {% endif %}
                </div>
                <div id="div_mail_body_main" style="display: none !important;"></div>
            </div>

            <div class="modal-footer">
                <button type="button" class="btn btn-default" data-dismiss="modal">
                    {% translate 'Close' %}
                </button>
            </div>
        </div>
    </div>
</div>
